from .XPS_Q8_drivers import XPS


class XPSError(Exception):
    """
    Exception related to the Newport XPS system
    """

    pass


class SimpleXPS:
    def __init__(
        self,
        ip: str,
        port: int,
        group: str,
        positionner: str,
    ):
        """
        Parameters
        ----------
        ip: str
            IP address to use for communication with the XPS motion controller. ex: "192.168.0.254"
        port: int
            IP port for communication with the XPS. ex: 5001
        group: str
            name of the group to control. ex: "Group2"
        positionner: str
            name of the positionner. ex: "Pos"
        """

        # init the wrapper given by Newport and some attributes
        self.xps = XPS()  # Instantiate the driver from Newport

        # required to connect via TCP/IP
        self._ip = ip
        self._port = port
        self.socket_id = -1

        # Definition of the stage
        self._group = group
        self._positioner = positionner
        self._full_positionner_name = f"{group}.{positionner}"

        # Some required initialisation steps
        self._init_commands()

    def _init_commands(self):
        """
        Runs some initial commands : connect to the XPS server, group kill, group intialize, move home.
        Some configs could be added here as well
        """
        self.socket_id = self.xps.TCP_ConnectToServer(
            self._ip, self._port, 5
        )  # 5s timeout
        # Check connection passed
        if self.socket_id == -1:
            raise XPSError("XPS_Q8 connection failed. Check ip address and port.")
        else:
            # Group kill to be sure
            [error_code, return_string] = self.xps.GroupKill(
                self.socket_id, self._group
            )
            if error_code != 0:
                self.display_error_and_close(error_code, "GroupKill")

            # Initialize
            [error_code, return_string] = self.xps.GroupInitialize(
                self.socket_id, self._group
            )
            if error_code != 0:
                self.display_error_and_close(error_code, "GroupInitialize")

            # Home search
            self.move_home()

            # Definition of the MotionDone trigger
            # [error_code, return_string] = self.xps.EventExtendedConfigurationTriggerSet(self.socket_id, 'MotionDone',0,0,0,0)
            # if (error_code != 0):
            #     self.display_error_and_close(error_code, 'EventExtendedConfigurationTriggerSet')
            # [error_code, return_string] = self.xps.EventExtendedConfigurationActionSet(self.socket_id, , 0, 0, 0, 0)
            # if (error_code != 0):
            #     self.display_error_and_close(error_code, 'EventExtendedConfigurationActionSet')

    def check_connected(self):
        """Returns true if the connection was successful, else false."""
        return self.socket_id != -1

    def display_error_and_close(self, error_code, api_name):
        """Method to recover an error string based on an error code. Closes the TCPIP connection afterwards"""
        if (error_code != -2) and (error_code != -108):
            [error_code, error_string] = self.xps.ErrorStringGet(
                self.socket_id, error_code
            )
            if error_code != 0:
                raise XPSError(f"{api_name} : ERROR {error_code}")
            else:
                raise XPSError(f"{api_name} : {error_string}")
        else:
            if error_code == -2:
                raise XPSError(f"{api_name} : TCP timeout")
            if error_code == -108:
                raise XPSError(
                    f"{api_name} : The TCP/IP connection was closed by an administrator"
                )
        self.close_tcpip()

    def close_tcpip(self):
        """Call the method to close the socket."""
        self.xps.TCP_CloseSocket(self.socket_id)

    def get_position(self):
        """Returns current the position"""
        if self.check_connected():
            [error_code, current_position] = self.xps.GroupPositionCurrentGet(
                self.socket_id, self._full_positionner_name, 1
            )
            if error_code != 0:
                self.display_error_and_close(error_code, "GroupPositionCurrentGet")
            else:
                return float(current_position)
        else:
            raise XPSError("XPS connection failed")

    def move_absolute(self, value):
        """Moves the stage to the position value."""
        if self.check_connected():
            [error_code, return_string] = self.xps.GroupMoveAbsolute(
                self.socket_id, self._full_positionner_name, [value]
            )
            if error_code != 0:
                self.display_error_and_close(error_code, "GroupMoveAbsolute")
        else:
            raise XPSError("XPS connection failed")

    def move_relative(self, value):
        """Moves the stage to value relative to it's current position."""
        if self.socket_id != -1:
            [error_code, return_string] = self.xps.GroupMoveRelative(
                self.socket_id, self._full_positionner_name, [value]
            )
            if error_code != 0:
                self.display_error_and_close(error_code, "GroupMoveRelative")
        else:
            raise XPSError("XPS connection failed")

    def move_home(self):
        """Moves the stage to it's home"""
        if self.check_connected():
            [error_code, return_string] = self.xps.GroupHomeSearch(
                self.socket_id, self._group
            )
            if error_code != 0:
                self.display_error_and_close(error_code, "GroupHomeSearch")
        else:
            raise XPSError("XPS connection failed")

    def set_group(self, group: str):
        """
        Sets the group to control with the plugin

        Args:
            group: Name of the group
        """
        self._group = group
        self._full_positionner_name = f"{group}.{self._positionner}"

    def set_positionner(self, positionner: str):
        """
        Sets the positionner to control with the plugin

        Args:
            positionner: Name of the positionner
        """
        self._positionner = positionner
        self._full_positionner_name = f"{self._group}.{positionner}"

    def set_ip(self, ip: str):
        """
        Sets the IP address to use for TCPIP communication with the XPS motion controller.

        Args:
            ip: IP address of the XPS device
        """
        self._ip = ip

    def set_port(self, port: int):
        """
        Sets the port to use for TCPIP communication with the XPS motion controller.

        Args:
            port: port to use
        """
        self._port = port

    def retry_connection(self):
        """
        Closes the existing TCPIP connection and tries to reconnect
        """
        self.close_tcpip()
        self._init_commands()



# -*- coding: utf-8 -*-
"""
Created on Mon Jan  9 14:59:52 2023

@author: weber
"""

import pyvisa
import numpy as np


class SerialBase(object):

    def __init__(self):
        super().__init__()
        self._controller = None
        self._VISA_rm = pyvisa.ResourceManager()
        self.com_ports = self.get_ressources()

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, to):
        self._timeout = to
        self._controller.timeout = to

    def get_ressources(self):
        infos=self._VISA_rm.list_resources_info()
        com_ports = [infos[key].alias for key in infos.keys()]
        return com_ports
    
    def init_communication(self, com_port, axis=1):
        if com_port in self.com_ports:
            self._controller = self._VISA_rm.open_resource(com_port)

            self._controller.data_bits = 8
            self._controller.stop_bits = pyvisa.constants.StopBits['one']
            self._controller.parity = pyvisa.constants.Parity['none']
            self.timeout = 2000
        

    def close_communication(self, axis=1):
        self._controller.close()
        self._VISA_rm.close() 
        
    def get_controller_infos(self, axis=1):
        self._write_command(f'{axis}ID?')
        return self._get_read()

    def _query(self, command):
        ret = self._controller.query(command)
        return ret

    def _write_command(self, command):
        self._controller.write(command)

    
    def _get_read(self):
        self._controller.timeout = 50
        info = ''
        try:
            while True:
                info += self._controller.read()+'\n'
        except pyvisa.errors.VisaIOError as e:
            pass
        self._controller.timeout = self._timeout
        return info
    
    
    def read(self):
        return self._get_read()
    
    def move_axis(self, move_type='ABS', axis=1, pos=0.):
        if move_type == 'ABS':
            ret = self._write_command(f'{axis}PA{pos}')

        elif move_type == 'REL':
            ret = self._write_command(f'{axis}PR{pos}')
        else:
            raise Exception('{:s} is not a valid displacement type'.format(move_type))
        return ret

    def get_position(self, axis=1):
        raise NotImplementedError


    def get_velocity_max(self, axis=1):
        raise NotImplementedError

    def set_velocity(self, velocity, axis=1):
        self._write_command(f'{axis}VA{velocity}')


    def move_home(self, axis=1):
        raise NotImplementedError

    def stop_motion(self, axis=1):
        self._write_command(f'{axis}ST')


from pymodaq.control_modules.move_utility_classes import (
    DAQ_Move_base,
    comon_parameters_fun,
    main,
    DataActuatorType,
    DataActuator,
)  # common set of parameters for all actuators
from pymodaq.utils.daq_utils import (
    ThreadCommand,
)  # object used to send info back to the main thread
from pymodaq.utils.parameter import Parameter
from pymodaq_plugins_newport.hardware.xps_q8_simplified import (
    SimpleXPS,
    XPSError,
)


class DAQ_Move_XpsQ8(DAQ_Move_base):
    """Instrument plugin class for Newport XPS Q8 Motion Controller.

    This object inherits all functionalities to communicate with PyMoDAQ’s DAQ_Move module through inheritance via
    DAQ_Move_base. It makes a bridge between the DAQ_Move module and the Python wrapper of a particular instrument.

    This plugin was tested on a Newport XPS Q8, using PyMoDAQ 2.4.2 on Windows 11.
    No proprietary driver is needed, the communication with the instrument is done through TCP/IP.
    This plugin may also be compatible with newer controllers (Newport XPS D series),
    please notify the author(s) if you have tested this plugin on a different model so it can be added
    to the list of compatible controllers.

    Attributes:
    -----------
    controller: object
        The particular object that allow the communication with the hardware, in general a python wrapper around the
         hardware library.
    """

    _controller_units = "mm"
    is_multiaxes = False
    _axis_names = ["Axis1"]
    _epsilon = 600e-6
    data_actuator_type = DataActuatorType["DataActuator"]

    params = [
        {
            "title": "XPS IP address :",
            "name": "xps_ip_address",
            "type": "str",
            "value": "192.168.0.254",
        },  # IP address of my system
        {
            "title": "XPS Port :",
            "name": "xps_port",
            "type": "int",
            "value": 5001,
        },  # Port of my system, should be the same for others ?
        {
            "title": "Group :",
            "name": "group",
            "type": "str",
            "value": "Group2",
        },  # Group to be moved
        {
            "title": "Positionner :",
            "name": "positionner",
            "type": "str",
            "value": "Pos",
        },  # positionner to be moved
    ] + comon_parameters_fun(is_multiaxes, axis_names=_axis_names, epsilon=_epsilon)

    def ini_attributes(self):
        self.controller: SimpleXPS | None = None

    def get_actuator_value(self):
        """Get the current value from the hardware with scaling conversion.

        Returns
        -------
        float: The position obtained after scaling conversion.
        """
        pos = DataActuator(data=self.controller.get_position())
        pos = self.get_position_with_scaling(pos)
        return pos

    def close(self):
        """Terminate the communication protocol"""
        if self.controller is not None:  # There's nothing to close otherwise
            self.controller.close_tcpip()

    def commit_settings(self, param: Parameter):
        """Apply the consequences of a change of value in the detector settings

        Parameters
        ----------
        param: Parameter
            A given parameter (within detector_settings) whose value has been changed by the user
        """
        if param.name() == "xps_ip_address":
            self.controller.set_ip(param.value())
        elif param.name() == "xps_port":
            self.controller.set_port(param.value())
        elif param.name() == "group":
            self.controller.set_group(param.value())
        elif param.name() == "positionner":
            self.controller.set_positionner(param.value())
        else:
            pass

    def ini_stage(self, controller=None):
        """Actuator communication initialization

        Parameters
        ----------
        controller: (object)
            custom object of a PyMoDAQ plugin (Slave case). None if only one actuator by controller (Master case)

        Returns
        -------
        info: str
        initialized: bool
            False if initialization failed otherwise True
        """

        try:
            new_controller = SimpleXPS(
                ip=self.settings["xps_ip_address"],
                port=self.settings["xps_port"],
                group=self.settings["group"],
                positionner=self.settings["positionner"],
            )
        except XPSError as e:
            initialized = False
            info = f"{e}"
            return info, initialized
        else:
            self.controller = self.ini_stage_init(
                old_controller=controller,
                new_controller=new_controller,
            )

        initialized = self.controller.check_connected()
        # here 'initialized' should always be True, as any error would have been caught above
        info = "XPS controller initialization"
        return info, initialized

    def move_abs(self, value: DataActuator):
        """Move the actuator to the absolute target defined by value

        Parameters
        ----------
        value: (float) value of the absolute target positioning
        """
        value = self.check_bound(
            value
        )  # if user checked bounds, the defined bounds are applied here
        self.target_value = value
        value = self.set_position_with_scaling(
            value
        )  # apply scaling if the user specified one
        self.emit_status(ThreadCommand("Update_Status", ["move_absolute command sent"]))
        try:
            self.controller.move_absolute(value.value())
        except XPSError as e:
            self.emit_status(ThreadCommand("Update_Status", [f"{e}"]))

    def move_rel(self, value: DataActuator):
        """Move the actuator to the relative target actuator value defined by value

        Parameters
        ----------
        value: (float) value of the relative target positioning
        """
        value = self.check_bound(self.current_position + value) - self.current_position
        self.target_value = value + self.current_position
        value = self.set_position_relative_with_scaling(value)

        self.emit_status(ThreadCommand("Update_Status", ["move_relative command sent"]))
        try:
            self.controller.move_relative(value.value())
        except XPSError as e:
            self.emit_status(ThreadCommand("Update_Status", [f"{e}"]))

    def move_home(self):
        """Call the reference method of the controller"""
        self.emit_status(ThreadCommand("Update_Status", ["moved_home command sent"]))
        try:
            self.controller.move_home()
        except XPSError as e:
            self.emit_status(ThreadCommand("Update_Status", [f"{e}"]))

    def stop_motion(self):
        """NOT IMPLEMENTED --- Stop the actuator and emits move_done signal"""

        ## Not possible to implement with this system as far as I'm aware.
        raise NotImplementedError


if __name__ == "__main__":
    main(__file__)
